@@ -3,7 +3,6 @@ import time
 from rest_framework import authentication
 from rest_framework import exceptions
 from django.conf import settings
-from django.shortcuts import get_object_or_404
 from django.contrib.auth import get_user_model
 from codebehind.models import UserSecret
 
@@ -19,6 +18,12 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 	# Authorization: X-HMAC-SHA256 Credential=USERNAME SignedHeaders=content-type;host;x-header Signature=CALCULATED_SIGNATURE
 	"""
 
+	def quote(self, text):
+		try:
+			return urllib.parse.quote(text)
+		except:
+			return urllib.quote(text)
+
 	def remove_prefix(self, text, prefix):
 		return text[text.startswith(prefix) and len(prefix):]
 
@@ -31,8 +36,8 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 		for header in request.META:
 
 			if regex_http.match(header):
-				strip_header = self.remove_prefix(header, "HTTP_").replace("_","-").lower()
-				request_headers[strip_header] = request.META[header]
+				strip_header = self.remove_prefix(header, "HTTP_").replace("_","-").lower().strip()
+				request_headers[strip_header] = self.trim_header_value(request.META[header])
 		return request_headers
 
 	def get_valid_headers(self, keys_of_valid_headers, headers):
@@ -42,7 +47,7 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 		keys = keys_of_valid_headers.split(";")
 		valid_headers = {}
 		for key in keys:
-			if headers.has_key(key):
+			if key in headers:
 				valid_headers[key] = headers[key]
 		return valid_headers
 
@@ -50,8 +55,16 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 		"""
 		this will do the hmac signing
 		"""
+		if isinstance(key, str):
+			key = key.encode('utf-8')
+		if isinstance(msg, str):
+			msg = msg.encode('utf-8')
+
 		return hmac.new(key, msg, hashlib.sha256)
 
+	def trim_header_value(self, text):
+		text = re.sub(r" +", " ", text)
+		return text.strip()
 
 	def get_signature_key(self, key, timestamp):
 		"""
@@ -66,29 +79,48 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 		"""
 		return sorted(dict.items())
 
+	def url_encode(self, text):
+		return self.quote(text)
+
 	def get_canonical_param(self, dict):
 		"""
-		This will sort(alphabetically regardless if it is uppercase or lowercase) and convert the params like query string or form data
-		into & separated values
+		- sorted by key
+		- convert value to string
+		- separated by =
+		- uri_encoded key
+		- uri_encoded value
+		
+		url_encoded(key)=url_encoded(value)
 
 		eg. age=20&name=kel&x-var=32
 		"""
-		if dict == None or len(dict) == 0:
-			return ''
-		return '&'.join("%s=%s" % (key,urllib.quote(val)) for (key,val) in self.sort_dict(dict))
+		try:
+			if dict == None or len(dict) == 0:
+				return ''
+			return '&'.join("%s=%s" % (self.url_encode(key), self.url_encode(str(val))) for (key,val) in self.sort_dict(dict))
+		except:
+			raise
 
 	def get_canonical_headers(self, headers):
 		"""
-		will sort the header name alpabetically and append \n
+		- header name and value is separated by colon
+		- reduced consecutive white spaces into single white space
+		- trimmed leading and trailing in header name
+		- append \n after every header value
+		
 		name:value\n
 
-		eg. content-type:application/json\nlength:100\n
+		eg. 
+		content-type:application/json\n
+		length:100\n
+
 		"""
-		return ''.join("%s:%s\n" % (key,urllib.quote(val)) for (key,val) in self.sort_dict(headers))
+		return ''.join("%s:%s\n" % (key, str(val)) for (key,val) in self.sort_dict(headers))
 		
 	def get_signed_headers(self, headers):
 		"""
-		semi-colon-separated header name(sorted alphabetically)
+		- only header name separate by semi colon.
+		eg.
 		content-type;length;
 		"""
 		return ';'.join("%s" % (key) for (key,val) in self.sort_dict(headers))
@@ -112,6 +144,10 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 		canonical_headers = self.get_canonical_headers(headers)
 		signed_headers = self.get_signed_headers(headers)
 		canonical_payload = self.get_canonical_param(payload)
+		
+		if isinstance(canonical_payload, str):
+			canonical_payload = canonical_payload.encode('utf-8')
+
 		canonical_payload_hash = hashlib.sha256(canonical_payload).hexdigest()
 
 		canonical_request = "%s\n%s\n%s\n%s\n%s\n%s" % (
@@ -122,6 +158,8 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 		 	signed_headers, 			 	
 		 	canonical_payload_hash)
 
+		canonical_request = canonical_request.encode('utf-8') 
+
 		canonical_request_hash = hashlib.sha256(canonical_request).hexdigest()
 
 		string_to_sign = "%s\n%s\n%s" % (
@@ -157,69 +195,92 @@ class CodeBehindAuthentication(authentication.BaseAuthentication):
 		"""
 
 		try:
+			cb_config = {}
+
+			if hasattr(settings, 'CODEBEHIND_CONFIG'):
+				cb_config = settings.CODEBEHIND_CONFIG
+
+			algorithm = cb_config.get("ALGORITHM", "X-HMAC-256")
+			signing_query_prefix = cb_config.get("SIGNING_QUERY_PREFIX","X")
+			validity_duration = cb_config.get("VALIDITY_DURATION", 15 * 60)
+
+			request_algorithm = ""
+			request_credential = ""
+			request_timestamp = ""
+			request_signed_headers = ""
+			request_signature = ""
+			request_qparams = request.query_params
+
 			request_headers = self.get_request_headers(request)
 
-			request_auth_header = request_headers.get("authorization", "")
+			if not request_headers.get("authorization") == None:
+				# if has authorization header then use this method
+				request_auth_header = request_headers.get("authorization", "")
+				auth_header_info = request_auth_header.split(" ")
+				if not len(auth_header_info) == 4:
+					return None
+				
+				request_algorithm = auth_header_info[0]
+				request_credential = auth_header_info[1].split("=")[1]
+				request_signed_headers = auth_header_info[2].split("=")[1]
+				request_valid_headers = self.get_valid_headers(request_signed_headers, 
+					request_headers)
+				request_signature = auth_header_info[3].split("=")[1]
+				request_timestamp = request_headers.get('x-timestamp')
 
-			auth_header_info = request_auth_header.split(" ")
-
-			if not len(auth_header_info) == 4:
-				return None
-
-
-			request_algorithm = auth_header_info[0]
-
-			algorithm = "X-HMAC-256"
-			if hasattr(settings, 'CODEBEHIND_ALGORITHM'):
-				algorithm = settings.CODEBEHIND_ALGORITHM
+			elif not request_qparams.get("%s-Algorithm" % signing_query_prefix) == None:
+				# if signed using query parameters
+				request_algorithm = request_qparams.get("%s-Algorithm" % signing_query_prefix)
+				request_credential = request_qparams.get("%s-Credential" % signing_query_prefix)
+				request_timestamp = request_qparams.get("%s-Timestamp" % signing_query_prefix)
+				request_signed_headers = request_qparams.get("%s-SignedHeaders" % signing_query_prefix)
+				request_valid_headers = self.get_valid_headers(request_signed_headers, 
+					request_headers)
+				request_signature = request_qparams.get("%s-Signature" % signing_query_prefix)
+				regex_signing_qparams = re.compile(r'^(%s-.+)$' % signing_query_prefix)
+				request_qparams = {key: value for key, value in request_qparams.items() if not regex_signing_qparams.match(key)}
+				
 
+			if not request_algorithm:
+				return (None)
+				
 			if not algorithm == request_algorithm:
-				return None
+				raise exceptions.ParseError("Algorithm is invalid.")
 
-			request_credential = auth_header_info[1]
-			request_username = request_credential.split("=")[1]
-			request_signed_headers = auth_header_info[2].split("=")[1]
-			request_valid_headers = self.get_valid_headers(request_signed_headers, 
-				request_headers)
-			request_signature = auth_header_info[3].split("=")[1]
-			
 			request_payload = request.data
 
 			canonical_uri = request.META.get('PATH_INFO')
-			request_qparams = request.query_params
 			content_type = request.content_type
 			request_method = request.method
-			request_timestamp = request_headers.get('x-timestamp')
-			user = get_object_or_404(get_user_model(), username='admin')
-			user_secret_key = user.secret
+			
+			user_q = get_user_model().objects.filter(username=request_credential)
+			if user_q.exists():
+				user = user_q.first()
+			else:
+				raise exceptions.ParseError("Credential is invalid.")
+
 			server_time_unix = float(time.time())
 
-			if abs(server_time_unix - float(request_timestamp)) > 60 * 2:
-				raise exceptions.AuthenticationFailed('Invalid request time!')
+			if abs(server_time_unix - float(request_timestamp)) > validity_duration:
+				raise exceptions.AuthenticationFailed('Invalid request time.')
 
 			access_key = user.username
 			signature = self.get_signature(
 				access_key, 
 				str(user.secret.key), 
-				algorithm, 
+				request_algorithm, 
 				request_method, 
 				canonical_uri,
 				request_timestamp, 
 				request_qparams,
 				request_valid_headers,
 				request_payload)
-
-			auth_header = self.get_auth_header(
-				access_key,
-				request_algorithm,
-				request_signed_headers,
-				signature)
-
+			
 			if not signature == request_signature:
-				raise exceptions.AuthenticationFailed('Signature dont matched!')
+				raise exceptions.AuthenticationFailed('Signature don\'t matched.')
 			
 			return (user, None)
 
-		except Exception:
-			raise exceptions.AuthenticationFailed('Invalid signature input!')
-
+		except:
+			raise
+		
